
from funcy import rcompose, ignore, fallback, mapcat
from .node import Node
from .schemas import media_schema
from modeller import Model
from .common import get_image_url, get_manifest, get_video_url
import traceback





class Media(Model, Node):

    def _on_init(self):
        try:
            self._validate()
        except Exception as e:
            print('ERROR in validation for Media:')
            print()
            print(str(e))
            print()
            print(self._yaml())
            print()

    _schema = media_schema

    __repr__ = lambda self: f'Media(url={self.url})'

    # id = property(lambda self: self.pk)

    url = property(lambda self: url_from_id(self.pk))

    # for image data, for video posts returns the thumbnail
    images = property(lambda self:  \
        list(map(get_image_url, self['carousel_media'])) if self['carousel_media'] \
        else [get_image_url(self)] or
        []
    )

    # for videos, it is a MDP, if it is long it needs to be recomposed
    mpd = property(lambda self:  \
        list(map(get_manifest, self['carousel_media'] or [])) if self['carousel_media'] \
        else [get_manifest(self)] or
        []
    )

    videos = property(lambda self:  \
        list(map(get_video_url, self['carousel_media'] or [])) if self['carousel_media'] \
        else [get_video_url(self)] or
        []
    )




    # def fallback(*args):
    #     first = lambda arr: arr[1:] if len(arr) > 0 else lambda: None
    #     rest = lambda arr: arr[:1]if len(arr) > 0 else []
    #     return  first(args)() or fallback(rest(args))

    _usertags = property(lambda self: fallback(
        lambda : self['usertags']['in']['user'],
        lambda : list(map(lambda x: x['user'], self['usertags']['in'])),
        lambda : list(map(lambda x: x['user'], self['caption']['usertags']['in'])),
        lambda: list(mapcat(lambda data: map(lambda x: x['user'], data['usertags']['in']), self['carousel_media'] or [])),
        lambda: list(mapcat(lambda data: map(lambda x: x['user'], data['caption']['usertags']['in']), self['carousel_media'] or [])),
        lambda: []
    ))

    # geotag = property(lambda self: self['location'])









def id_from_url(link):
    if 'instagram.com/p/' not in link:
        # self.logger.error('Unexpected link')
        return False
    link = link.split('/')
    code = link[link.index('p') + 1]

    alphabet = {
        '-': 62, '1': 53, '0': 52, '3': 55, '2': 54, '5': 57, '4': 56,
        '7': 59, '6': 58, '9': 61, '8': 60, 'A': 0, 'C': 2, 'B': 1,
        'E': 4, 'D': 3, 'G': 6, 'F': 5, 'I': 8, 'H': 7, 'K': 10, 'J': 9,
        'M': 12, 'L': 11, 'O': 14, 'N': 13, 'Q': 16, 'P': 15, 'S': 18,
        'R': 17, 'U': 20, 'T': 19, 'W': 22, 'V': 21, 'Y': 24, 'X': 23,
        'Z': 25, '_': 63, 'a': 26, 'c': 28, 'b': 27, 'e': 30, 'd': 29,
        'g': 32, 'f': 31, 'i': 34, 'h': 33, 'k': 36, 'j': 35, 'm': 38,
        'l': 37, 'o': 40, 'n': 39, 'q': 42, 'p': 41, 's': 44, 'r': 43,
        'u': 46, 't': 45, 'w': 48, 'v': 47, 'y': 50, 'x': 49, 'z': 51,
        }

    result = 0
    for char in code:
        result = result * 64 + alphabet[char]
    return result


def url_from_id(id):
    id = int(id)
    alphabet = {
        '-': 62, '1': 53, '0': 52, '3': 55, '2': 54, '5': 57, '4': 56,
        '7': 59, '6': 58, '9': 61, '8': 60, 'A': 0, 'C': 2, 'B': 1,
        'E': 4, 'D': 3, 'G': 6, 'F': 5, 'I': 8, 'H': 7, 'K': 10, 'J': 9,
        'M': 12, 'L': 11, 'O': 14, 'N': 13, 'Q': 16, 'P': 15, 'S': 18,
        'R': 17, 'U': 20, 'T': 19, 'W': 22, 'V': 21, 'Y': 24, 'X': 23,
        'Z': 25, '_': 63, 'a': 26, 'c': 28, 'b': 27, 'e': 30, 'd': 29,
        'g': 32, 'f': 31, 'i': 34, 'h': 33, 'k': 36, 'j': 35, 'm': 38,
        'l': 37, 'o': 40, 'n': 39, 'q': 42, 'p': 41, 's': 44, 'r': 43,
        'u': 46, 't': 45, 'w': 48, 'v': 47, 'y': 50, 'x': 49, 'z': 51,
        }
    result = ''
    while id:
        id, char = id // 64, id % 64
        result += list(alphabet.keys())[list(alphabet.values()).index(char)]
    return 'https://instagram.com/p/' + result[::-1] + '/'


Media.id_from_url = id_from_url


from typing import List
from funcy import rcompose, mapcat
from ..nodes import User
from .common import decorate, cycled_api_call


@decorate(accepts=User, returns=User)
def following(bot, nodes,  args) -> List[User]:

    amount = args.get('amount') or 1

    pack_user = lambda item: User(**item)

    process = rcompose(
        lambda user: user.pk,
        lambda id: cycled_api_call(amount, bot, bot.api.user_following, dict(user_id=id, **args.get('query', {}),), 'users'),
        lambda gen: map(pack_user, gen)
    )

    result = mapcat(process, nodes)

    return result, {}



import codecs
# from http.cookies import BaseCookie
from ..api.instagram_private_api.compat import compat_cookiejar
def to_json(python_object):
    if isinstance(python_object, bytes):
        return {'__class__': 'bytes',
                '__value__': codecs.encode(python_object, 'base64').decode()}
    raise TypeError(repr(python_object) + ' is not JSON serializable')


def from_json(json_object):
    if '__class__' in json_object and json_object['__class__'] == 'bytes':
        return codecs.decode(json_object['__value__'].encode(), 'base64')
    return json_object


def serialize_cookie(cookie):
    obj = {}
    for name in (
        "version", "name", "value", "port", "port_specified",
        "domain", "domain_specified", "domain_initial_dot",
        "path", "path_specified",
        "secure", "expires", "discard", "comment", "comment_url",
    ):
        obj[name] = getattr(cookie, name)
    return obj

def serialize_cookie_jar(jar, ):
    result={}
    for k, v in jar.items():
        if isinstance(v, compat_cookiejar.Cookie):
            result[k] = serialize_cookie(v)
        elif isinstance(v, dict):
            result.update(serialize_cookie_jar(v, ))
    return result

def deserialize_cookie(cookie):
    obj = {}
    for name in (
        "version", "name", "value", "port", "port_specified",
        "domain", "domain_specified", "domain_initial_dot",
        "path", "path_specified",
        "secure", "expires", "discard", "comment", "comment_url", "rest"
    ):
        obj[name] = cookie.get(name)
    return compat_cookiejar.Cookie(**obj)

def deserialize_cookie_jar(jar):
    result={}
    for k, v in jar.items():
        if 'name' in v and 'version' in v:
            result[k] = deserialize_cookie(v)
        elif isinstance(v, dict):
            result[k] = deserialize_cookie_jar(v, )
        else:
            result[k] = v
    return result



from .common import decorate
from ..nodes import User
from datetime import datetime
from .common import tap
from ..bot import Bot
from funcy import raiser, rcompose




@decorate(accepts=User, returns=User)
def follow(bot: Bot, nodes,  args):

    max = float(args['max']) if 'max' in args else float('inf')
    count = 0
    events = []

    def increment():
        nonlocal count
        count += 1
        
    def add_event(node: User):
        events.append({
            'type': 'follow',
            'metadata': bot.metadata,
            'args': {},
            'node': {
                'type': 'user',
                'username': node.username,
            },
            'timestamp': int(datetime.utcnow().timestamp())
        })
        return node

    stop = raiser(StopIteration)

    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        lambda node: follow_user(node, bot=bot) \
            if node else None,
        lambda x: tap(x, increment) if x else None,
        lambda x: add_event(x) if x else None,
    )


    followed = map(process, nodes)
    followed = filter(lambda x: x, followed)

    return followed, { 'events': events }



def follow_user(user, bot):
    bot.api.friendships_create(user.pk)
    bot.total['follows'] += 1
    bot.logger.info(f'followed {user}')
    bot.sleep('follow')
    return user
